import libximc as ximc
from ctypes import byref, c_int, c_float


WAIT_FOR_STOP_MS = 10


class Actuator():
    def __init__(self, device_uri):
        self._device_uri = device_uri
        self._device_id = None
        self._unit_converter = None

        if "xi-emu" in self._device_uri:
            self.setup_virtual_controller()

    def setup_virtual_controller(self):
        self.connect()

        engine_settings = ximc.engine_settings_t()
        ximc.lib.get_engine_settings(self._device_id, byref(engine_settings))
        engine_settings.StepsPerRev = 200
        ximc.lib.set_engine_settings(self._device_id, byref(engine_settings))

        feedback_settings = ximc.feedback_settings_t()
        ximc.lib.get_feedback_settings(self._device_id, byref(feedback_settings))
        feedback_settings.CountsPerTurn = 20000
        ximc.lib.set_feedback_settings(self._device_id, byref(feedback_settings))

        sync_out_settings = ximc.sync_out_settings_t()
        ximc.lib.get_sync_out_settings(self._device_id, byref(sync_out_settings))
        sync_out_settings.SyncOutPeriod = 100
        ximc.lib.set_sync_out_settings(self._device_id, byref(sync_out_settings))

        self.disconnect()

    def connect(self):
        self._device_id = ximc.lib.open_device(self._device_uri.encode())

        if self._device_id > 0:
            print("Device with URI {} successfully opened".format(self._device_uri))
        else:
            raise RuntimeError("Failed to open device with URI", self._device_uri)

        # One turn = 200 steps = 360 degrees
        engine_settings = ximc.engine_settings_t()
        ximc.lib.get_engine_settings(self._device_id, byref(engine_settings))
        step_to_degr_conversion_coeff = 360.0 / engine_settings.StepsPerRev
        self._unit_converter = ximc.calibration_t()
        self._unit_converter.MicrostepMode = ximc.MicrostepMode.MICROSTEP_MODE_FRAC_256
        self._unit_converter.A = step_to_degr_conversion_coeff

    def disconnect(self):
        ximc.lib.close_device(byref(c_int(self._device_id)))

    def get_position(self):
        position_struct = ximc.get_position_calb_t()
        ximc.lib.get_position_calb(self._device_id, byref(position_struct), byref(self._unit_converter))
        angle = position_struct.Position
        return angle

    def wait_for_stop(self, ms: int):
        ximc.lib.command_wait_for_stop(self._device_id, ms)

    def move_to_position(self, angle: float):
        ximc.lib.command_move_calb(self._device_id, c_float(angle), byref(self._unit_converter))
        self.wait_for_stop(WAIT_FOR_STOP_MS)

    def movr_to_position(self, angle: float):
        ximc.lib.command_movr_calb(self._device_id, c_float(angle), byref(self._unit_converter))
        self.wait_for_stop(WAIT_FOR_STOP_MS)

    def zero(self):
        ximc.lib.command_zero(self._device_id)

    def right(self):
        ximc.lib.command_right(self._device_id)

    def left(self):
        ximc.lib.command_left(self._device_id)

    def sstp(self):
        ximc.lib.command_sstp(self._device_id)

    def get_measurements_per_revolution(self):
        feedback_settings = ximc.feedback_settings_t()
        ximc.lib.get_feedback_settings(self._device_id, byref(feedback_settings))

        sync_out_settings = ximc.sync_out_settings_t()
        ximc.lib.get_sync_out_settings(self._device_id, byref(sync_out_settings))

        if sync_out_settings.SyncOutPeriod == 0:
            raise RuntimeError(
                "SyncOutPeriod = 0. Check XIMC motor controller profile!!!"
            )

        return feedback_settings.CountsPerTurn / sync_out_settings.SyncOutPeriod

    def check_ender_switch_button(self):
        status = ximc.status_t()
        ximc.lib.get_status(self._device_id, byref(status))
        return (status.GPIOFlags & ximc.GPIOFlags.STATE_LEFT_EDGE)
