import enum
import sys
import os
import platform
import subprocess
from typing import Optional
import csv
import glob
import re
from PyQt5.QtCore import pyqtSignal, pyqtSlot, QObject, QTimer

mount_path = r"/mnt"
file_directory_name = r"measurements"
media_check_frequency = 1  # Hz


class ExternalMediaState(enum.IntEnum):
    DISCONNECTED = 0
    CONNECTED = 1


class _ExternalMediaWatchdog(QObject):
    """
    Class to check presence of external media.
    """

    external_media_state_changed: pyqtSignal = pyqtSignal(ExternalMediaState)

    def __init__(self, check_frequency: Optional[float], external_media_path: Optional[str]):
        """
        :param check_frequency: frequency in Hz with which the connection to
        the external media is checked;
        :param external_media_path: path to directory with external media;
        :param external_media_name: name of external media;
        :param parent:
        """

        super().__init__()
        self.__external_media_name: str = None
        self.__external_media_path: str = external_media_path
        self.__timer: QTimer = QTimer(parent=self)
        self.__timer.timeout.connect(self.__on_timeout)
        self.__timer.setInterval(int(1000.0 / check_frequency))  # in ms
        self.__timer.setSingleShot(False)

    @pyqtSlot()
    def __on_timeout(self):
        media_was_connected = self.__external_media_name is not None
        self.__external_media_name = _find_external_media(self.__external_media_path)
        media_is_connected = self.__external_media_name is not None
        if media_is_connected != media_was_connected:
            self.external_media_state_changed.emit(ExternalMediaState.CONNECTED if media_is_connected else
                                                   ExternalMediaState.DISCONNECTED)

    @pyqtSlot()
    def start(self):
        self.__timer.start()

    @pyqtSlot()
    def stop(self):
        self.__timer.stop()

    def get_external_media_name(self):
        return self.__external_media_name


def _find_external_media(external_media_path: str) -> Optional[str]:
    """
    Function looks for new external media in given directory.
    :param external_media_path: path to directory with external media.
    :return: one of external media.
    """

    if external_media_path is None:
        return None

    dirs = os.listdir(external_media_path)

    if len(dirs) == 0:
        return None
    # Note: directories are not ordered
    for external_media_name in dirs:
        if is_mounted(os.path.join(external_media_path, external_media_name)):
            return external_media_name
    return None


def is_mounted(mount_path: str) -> bool:
    """
    Function checks if there is required mounted path in OS.
    :param mount_path: required mounted path.
    :return: True if path is mounted.
    """

    if sys.platform != "linux":
        return True
    if platform.processor() == "x86_64":
        return True
    process = subprocess.Popen(["mount"], stdout=subprocess.PIPE)
    for line in iter(process.stdout.readline, ""):
        chunks = line.decode("utf8").split(" ")
        if len(chunks) < 3:
            return False
        if chunks[2] == mount_path:
            for a in os.listdir("/dev"):
                if os.path.join("/dev", a) == chunks[0]:
                    return True
    return False


class FileManager(QObject):
    external_media_state_changed: pyqtSignal = pyqtSignal(ExternalMediaState)

    def __init__(self):
        super().__init__()
        self.__external_media_path: str = mount_path
        self.__directory_name = file_directory_name
        self.__external_media_watchdog = _ExternalMediaWatchdog(check_frequency=media_check_frequency,
                                                                external_media_path=self.__external_media_path)
        self.__external_media_watchdog.external_media_state_changed.connect(self._on_external_media_state_changed)

    def start_watchdog(self):
        self.__external_media_watchdog.start()

    def get_filepath(self):
        filepath = os.path.join(
            self.__external_media_path, self.__external_media_watchdog.get_external_media_name(), self.__directory_name
        )
        return filepath

    def get_filename(self, path):
        names = glob.glob(path + '/*.csv')
        s = [int(re.findall(r"\d+", name)[0]) for name in names]
        if len(s) > 0:
            num = max(s)
        else:
            num = 0
        filename = str(num + 1) + '.csv'
        return filename

    def save_data(self, angles, res1, res2):
        filepath = self.get_filepath()
        if not os.path.exists(filepath):
            os.mkdir(filepath)

        filename = self.get_filename(filepath)
        file = os.path.join(filepath, filename)

        rows = list(zip(*[angles, res1, res2]))
        with open(file, 'w') as csvfile:
            writer = csv.writer(csvfile, dialect='excel')
            writer.writerow(['Angle degr', 'R bottom Ohm', 'R top Ohm'])
            writer.writerows(rows)
            os.fsync(csvfile)
        os.sync()
        return filename

    @pyqtSlot(ExternalMediaState)
    def _on_external_media_state_changed(self, state: ExternalMediaState):
        """
        Slot handles signal about changing of external media connection.
        :param state: new state of connection.
        """
        self.external_media_state_changed.emit(state)
