from PyQt5.QtCore import Qt
from PyQt5.uic import loadUi
from PyQt5 import QtGui
from PyQt5.QtCore import QObject, QTimer, pyqtSignal, pyqtSlot
from PyQt5.QtWidgets import QMainWindow, QWidget, QStackedWidget, QMessageBox, QPushButton, qApp, QGraphicsView
import numpy as np
from matplotlib import pyplot as plt
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from platform import system
from file_manager import FileManager, ExternalMediaState
from measurer import Measurer, MeasurementState, MeasurementStateSignal
from linear_error_calculation import calculate_r_nom_real, calculate_linear_error_and_angle_range


WINDOW_WIDTH = 1024
WINDOW_HEIGHT = 600

nominal_resistance_values = [100, 150, 220, 330, 470, 680, 1000, 1500, 2200, 3300, 4700, 6800, 10000, 15000]

ok_color = 'lightgreen'
error_color = 'pink'
default_style_sheet = 'QFrame{border-width: 1px; border-style: solid; border-color: black;}'
green_style_sheet = 'QFrame{background-color: ' + ok_color + \
    '; border-width: 1px; border-style: solid; border-color: black;}'
red_style_sheet = 'QFrame{background-color :' + error_color + \
    '; border-width: 1px; border-style: solid; border-color: black;}'

status_text_wait_style = '''QFrame{
    background: rgb(212, 212, 212);
    border-style: outset;
    border-width: 2px;
    border-radius: 5px;
    border-color: gray;
}'''
status_text_ok_style = '''QFrame{
    background: ''' + ok_color + ''';
    border-style: outset;
    border-width: 2px;
    border-radius: 5px;
    border-color: green;
}'''
status_text_error_style = '''QFrame{
    background:  ''' + error_color + ''';
    border-style: outset;
    border-width: 2px;
    border-radius: 5px;
    border-color: red;
}'''


class MeasurementSignalEmitter(QObject):
    measurement_state_signal = pyqtSignal(MeasurementStateSignal)


class MeasurementTimer(QTimer):
    def __init__(self):
        super().__init__()
        self.signal_emitter = MeasurementSignalEmitter()


class MainWindow(QMainWindow):
    def __init__(self, main_widget):
        super().__init__()
        self.setFixedHeight(WINDOW_HEIGHT)
        self.setFixedWidth(WINDOW_WIDTH)
        self.setCentralWidget(main_widget)

        if system() == 'Linux':
            self.setCursor(Qt.BlankCursor)


class GraphDisplay(QMessageBox):
    def __init__(self):
        super().__init__()
        w = 940
        h = 460
        super().setStyleSheet(
            "QLabel{min-width: " + str(w) + "px; min-height: " + str(h) + "; font: regular; font-size: 24px;} \
             QPushButton{ width:200px; height: 60px; font-size: 24px; }"
        )
        self.graph_view = QGraphicsView(parent=self)
        self.graph_view.setGeometry(0, 0, w, h)

        self.fig = plt.figure(figsize=(5, 1.5), dpi=72)
        self.ax1 = self.fig.add_subplot(211)
        self.ax2 = self.fig.add_subplot(212)
        self.fc = FigureCanvas(self.fig)
        self.fc.resize(w, h)
        self.fc.setParent(self.graph_view)

        if system() == 'Linux':
            self.setCursor(Qt.BlankCursor)

    def draw_graph(self, angles, resistance1, resistance2, resistances1, resistances2, deviation):
        self.ax1.cla()
        self.ax2.cla()
        self.ax1.set_xlabel("Угол, град.")
        self.ax1.set_ylabel("Сопротивление, Ом")
        self.ax1.set_xlim(0, 360)
        self.ax2.set_xlabel("Угол, град.")
        self.ax2.set_ylabel("Относ. отклонение, %")
        self.ax2.set_xlim(0, 360)
        if resistance1 != 0:
            try:
                resistances1_aligned, x1, y1 = calculate_linear_error_and_angle_range(
                    angles, resistances1, resistance1, return_data=True
                )
                self.ax1.plot(angles, resistances1_aligned, '.', label="Верх. вход (ADC 2)", color="C0")
                self.ax2.plot(x1, y1, '.', label="Верх. вход (ADC 2)", color="C0")
            except Exception:
                self.ax1.plot(angles, resistances1, '.', label="Верх. вход (ADC 2)", color="C0")
        if resistance2 != 0:
            try:
                resistances2_aligned, x2, y2 = calculate_linear_error_and_angle_range(
                    angles, resistances2, resistance2, return_data=True
                )
                self.ax1.plot(angles, resistances2_aligned, '.', label="Нижн. вход (ADC 1)", color="C1")
                self.ax2.plot(x2, y2, '.', label="Нижн. вход (ADC 1)", color="C1")
            except Exception:
                self.ax1.plot(angles, resistances2, '.', label="Нижн. вход (ADC 1)", color="C1")
        self.ax2.axhline(deviation, color='C3', linestyle="dashed", label="Допуск")
        self.ax2.axhline(-deviation, color='C3', linestyle="dashed")
        self.ax1.legend()
        self.ax2.legend()
        self.fc.draw()
        qApp.processEvents()

    def exec(self):
        super().exec_()


class StackedWidget(QStackedWidget):
    def __init__(self, resistances_widget, results_widget, deviation_widget, measurer: Measurer):
        super().__init__()
        self.addWidget(resistances_widget)
        self.addWidget(deviation_widget)
        self.addWidget(results_widget)
        self.measurer = measurer
        self.resistance1 = 100
        self.resistance2 = 0     # 0 means it is not used
        self.deviation = 0.5
        self.deviation_string = 'ПЗ'
        self.cur_res = 1
        self.button1 = self.widget(0).button_100
        self.button2 = self.widget(0).button_none

        self.button1.setChecked(True)
        self.widget(1).text_upper.setText("Верхняя секция: 100 Ом")
        self.widget(1).text_lower.setText("Нижняя секция: не используется")

        self.widget(0).button_100.clicked.connect(lambda: self.select_resistance(100, self.widget(0).button_100))
        self.widget(0).button_150.clicked.connect(lambda: self.select_resistance(150, self.widget(0).button_150))
        self.widget(0).button_220.clicked.connect(lambda: self.select_resistance(220, self.widget(0).button_220))
        self.widget(0).button_330.clicked.connect(lambda: self.select_resistance(330, self.widget(0).button_330))
        self.widget(0).button_470.clicked.connect(lambda: self.select_resistance(470, self.widget(0).button_470))
        self.widget(0).button_680.clicked.connect(lambda: self.select_resistance(680, self.widget(0).button_680))
        self.widget(0).button_1000.clicked.connect(lambda: self.select_resistance(1000, self.widget(0).button_1000))
        self.widget(0).button_1500.clicked.connect(lambda: self.select_resistance(1500, self.widget(0).button_1500))
        self.widget(0).button_2200.clicked.connect(lambda: self.select_resistance(2200, self.widget(0).button_2200))
        self.widget(0).button_3300.clicked.connect(lambda: self.select_resistance(3300, self.widget(0).button_3300))
        self.widget(0).button_4700.clicked.connect(lambda: self.select_resistance(4700, self.widget(0).button_4700))
        self.widget(0).button_6800.clicked.connect(lambda: self.select_resistance(6800, self.widget(0).button_6800))
        self.widget(0).button_10000.clicked.connect(lambda: self.select_resistance(10000, self.widget(0).button_10000))
        self.widget(0).button_15000.clicked.connect(lambda: self.select_resistance(15000, self.widget(0).button_15000))
        self.widget(0).button_none.clicked.connect(lambda: self.select_resistance(0, self.widget(0).button_none))
        self.widget(0).button_next.clicked.connect(lambda: self.setCurrentIndex(1))
        self.widget(0).button_upper.clicked.connect(lambda: self.set_resistor('upper'))
        self.widget(0).button_lower.clicked.connect(lambda: self.set_resistor('lower'))

        self.widget(1).button_5.clicked.connect(lambda: self.select_deviation(0.5))
        self.widget(1).button_45.clicked.connect(lambda: self.select_deviation(0.45))
        self.widget(1).button_4.clicked.connect(lambda: self.select_deviation(0.4))
        self.widget(1).button_measure.clicked.connect(self.measurement_start)
        self.widget(1).button_back.clicked.connect(lambda: self.setCurrentIndex(0))

        self.widget(2).button_back.clicked.connect(lambda: self.setCurrentIndex(0))
        self.widget(2).button_graph.clicked.connect(self.display_graph)
        self.widget(2).button_terminate.clicked.connect(self.measurer.terminate)
        self.widget(2).button_terminate.setEnabled(False)
        self.widget(2).button_again.clicked.connect(self.measurement_start)
        self.widget(2).progressBar.setVisible(False)
        self.widget(2).progressBar.setValue(0)

        # Measurement timer
        self.measurer.timer = MeasurementTimer()
        self.measurer.timer.setInterval(self.measurer.timeout_for_single_measurement)
        self.measurer.timer.timeout.connect(self.measurer.pick_measurements)
        self.measurer.timer.signal_emitter.measurement_state_signal.connect(self.on_measurement_state_changed)

        # Timer to register button press
        self.button_pressed = False
        self.button_timer = QTimer()
        self.button_timer.setInterval(100)
        self.button_timer.timeout.connect(self.check_hardware_start_button)
        self.button_timer.start()

        self.measurer.progressBar = self.widget(2).progressBar

        self.usb_connected = False
        if system() == 'Linux':
            self.file_manager = FileManager()
            self.file_manager.external_media_state_changed.connect(self.on_external_media_state_changed)
            self.file_manager.start_watchdog()

        self.results_dict1 = {}
        self.results_dict1['resistance_text'] = self.widget(2).resistance1
        self.results_dict1['li_ch'] = self.widget(2).li_ch1
        self.results_dict1['angle'] = self.widget(2).angle1
        self.results_dict1['message'] = "Ошибка при вычислении линейности верхнего резистора\n"
        self.results_dict2 = {}
        self.results_dict2['resistance_text'] = self.widget(2).resistance2
        self.results_dict2['li_ch'] = self.widget(2).li_ch2
        self.results_dict2['angle'] = self.widget(2).angle2
        self.results_dict2['message'] = "Ошибка при вычислении линейности нижнего резистора\n"

    def select_resistance(self, res, button):
        if self.cur_res == 1:
            self.resistance1 = res
            self.button1 = button
            if res == 0:
                self.widget(0).label_upper.setText("Не используется")
                self.widget(1).text_upper.setText("Верхняя секция: не используется")
            else:
                self.widget(0).label_upper.setText("Выбрано: {} Ом".format(res))
                self.widget(1).text_upper.setText("Верхняя секция: {} Ом".format(res))

        elif self.cur_res == 2:
            self.resistance2 = res
            self.button2 = button
            if res == 0:
                self.widget(0).label_lower.setText("Не используется")
                self.widget(1).text_lower.setText("Нижняя секция: не используется")
            else:
                self.widget(0).label_lower.setText("Выбрано: {} Ом".format(res))
                self.widget(1).text_lower.setText("Нижняя секция: {} Ом".format(res))

        if self.resistance1 == 0 and self.resistance2 == 0:
            self.widget(0).button_next.setEnabled(False)
        else:
            self.widget(0).button_next.setEnabled(True)

    def select_deviation(self, deviation):
        self.deviation = deviation
        if np.isclose(deviation, 0.5):
            self.deviation_string = 'ПЗ'
        elif np.isclose(deviation, 0.45):
            self.deviation_string = 'ОТК'
        elif np.isclose(deviation, 0.4):
            self.deviation_string = 'Цех'

    def set_resistor(self, type: str):
        if type == 'upper':
            self.cur_res = 1
            self.button1.setChecked(True)
        elif type == 'lower':
            self.cur_res = 2
            self.button2.setChecked(True)

    def display_graph(self):
        if len(self.measurer.measurements) == 0:
            self.display_warning_message('Нет измерений')
            return

        view = GraphDisplay()
        angles, resistances1, resistances2 = self.measurer.get_measurements()
        view.draw_graph(angles, self.resistance1, self.resistance2, resistances1, resistances2, self.deviation)
        view.exec()

    def gui_on_start(self):
        self.widget(2).button_back.setEnabled(False)
        self.widget(2).button_terminate.setEnabled(True)
        self.widget(2).button_again.setEnabled(False)
        self.widget(2).button_graph.setEnabled(False)
        self.widget(2).status_text.setStyleSheet(status_text_wait_style)
        self.widget(2).status_text.setText('Подождите...')

        if self.resistance1 != 0:
            self.widget(2).gen_label1.setText(f'{self.resistance1} Ом {self.deviation_string}')
        else:
            self.widget(2).gen_label1.setText('Не используется')
        if self.resistance2 != 0:
            self.widget(2).gen_label2.setText(f'{self.resistance2} Ом {self.deviation_string}')
        else:
            self.widget(2).gen_label2.setText('Не используется')

        self.widget(2).resistance1.setText("-")
        self.widget(2).resistance2.setText("-")
        self.widget(2).resistance1.setStyleSheet(default_style_sheet)
        self.widget(2).resistance2.setStyleSheet(default_style_sheet)
        self.widget(2).li_ch1.setText("-")
        self.widget(2).li_ch2.setText("-")
        self.widget(2).li_ch1.setStyleSheet(default_style_sheet)
        self.widget(2).li_ch2.setStyleSheet(default_style_sheet)
        self.widget(2).angle1.setText("-")
        self.widget(2).angle2.setText("-")
        self.widget(2).angle1.setStyleSheet(default_style_sheet)
        self.widget(2).angle2.setStyleSheet(default_style_sheet)

        self.widget(2).progressBar.setVisible(True)
        self.widget(2).progressBar.setValue(0)

        self.widget(2).data_saved_state.setText('Измерения не сохранены')

        self.setCurrentIndex(2)
        qApp.processEvents()

    def gui_on_finish(self):
        self.widget(2).button_back.setEnabled(True)
        self.widget(2).button_terminate.setEnabled(False)
        self.widget(2).button_again.setEnabled(True)
        self.widget(2).button_graph.setEnabled(True)
        self.widget(2).progressBar.setVisible(False)
        self.widget(2).progressBar.setValue(0)

    def measurement_start(self):
        self.gui_on_start()
        if self.measurer.virtual:
            self.measurer.ohmmeter.nominal_resistance1 = self.resistance1
            self.measurer.ohmmeter.nominal_resistance2 = self.resistance2
        self.measurer.start()

    @pyqtSlot(MeasurementStateSignal)
    def on_measurement_state_changed(self, state=MeasurementStateSignal):
        if state == MeasurementStateSignal.FINISHED:
            self.measurement_finish()
        if state == MeasurementStateSignal.TERMINATED:
            self.measurement_terminate()
        if state == MeasurementStateSignal.TIMEOUT:
            self.display_warning_message('Не удалось получить измерения.\nПроверьте, что мотор подключен.')

    def measurement_terminate(self):
        self.gui_on_finish()
        self.widget(2).status_text.setText('Прервано')

    def measurement_finish(self):
        self.gui_on_finish()
        angles, resistances1, resistances2 = self.measurer.get_measurements()

        # Сохранение измерений
        if self.usb_connected:
            try:
                filename = self.file_manager.save_data(angles, resistances1, resistances2)
                self.widget(2).data_saved_state.setText(f'Измерения не сохранены в {filename}')
            except Exception as exc:
                self.widget(2).data_saved_state.setText('Ошибка при сохранении')
                raise exc

        # Обработка измерений
        msg = ""
        error1, error2 = 0, 0
        if self.resistance1 != 0:
            error1, msg1 = self.analyze_measurements(angles, resistances1, self.resistance1, self.results_dict1)
            msg += msg1
        if self.resistance2 != 0:
            error2, msg2 = self.analyze_measurements(angles, resistances2, self.resistance2, self.results_dict2)
            msg += msg2

        success = (error1 < self.deviation) & (error2 < self.deviation)

        if success:
            self.widget(2).status_text.setText("Ок")
            self.widget(2).status_text.setStyleSheet(status_text_ok_style)
        else:
            self.widget(2).status_text.setText("Тест не пройден")
            self.widget(2).status_text.setStyleSheet(status_text_error_style)

        if msg != "":
            msg += "Проверьте заданные сопротивления и подключение к резисторам."
            self.display_warning_message(msg)

    def analyze_measurements(self, angles, measurements, resistance, gui_dict):
        ok_flag = True
        error = 0
        r_real = calculate_r_nom_real(measurements)
        gui_dict['resistance_text'].setText('{:.1f}'.format(r_real))
        if abs(r_real - resistance) / resistance <= 0.05:
            gui_dict['resistance_text'].setStyleSheet(green_style_sheet)
        else:
            ok_flag = False
            gui_dict['resistance_text'].setStyleSheet(red_style_sheet)

        msg = ''
        try:
            error, angle = calculate_linear_error_and_angle_range(angles, measurements, resistance)
            gui_dict['li_ch'].setText('{:.3f}'.format(error))
            gui_dict['angle'].setText('{:.1f}'.format(angle))
            if error < self.deviation:
                gui_dict['li_ch'].setStyleSheet(green_style_sheet)
            else:
                ok_flag = False
                gui_dict['li_ch'].setStyleSheet(red_style_sheet)

            if abs(angle - 330) <= 3:
                gui_dict['angle'].setStyleSheet(green_style_sheet)
            else:
                ok_flag = False
                gui_dict['angle'].setStyleSheet(red_style_sheet)
        except Exception as e:
            ok_flag = False
            msg = gui_dict['message'] + "\n" + str(e) + "\n"

        if not ok_flag:
            error = 10
        return (error, msg)

    @pyqtSlot(ExternalMediaState)
    def on_external_media_state_changed(self, state: ExternalMediaState):
        if state == ExternalMediaState.DISCONNECTED:
            self.usb_connected = False
            self.widget(2).usb_state.setText('Нет USB-носителя')
        elif state == ExternalMediaState.CONNECTED:
            self.usb_connected = True
            self.widget(2).usb_state.setText('USB-носитель подключен')

    def display_warning_message(self, msg_text):
        msg = QMessageBox()
        if system() == 'Linux':
            msg.setCursor(Qt.BlankCursor)
        msg.setFont(QtGui.QFont('MS Shell Dlg 2'))
        exit_btn = QPushButton("OK", msg)
        msg.addButton(exit_btn, QMessageBox.AcceptRole)
        msg.setStyleSheet(
            "QLabel{ font: regular; font-size: 24px;} QPushButton{ width:200px; height: 60px; font-size: 24px; }"
        )
        msg.setText("Ошибка")
        msg.setInformativeText(msg_text)
        msg.exec_()

    def check_hardware_start_button(self):
        if self.measurer.actuator.check_ender_switch_button() and self.button_pressed:
            # already reacted, ignore
            return

        if self.measurer.actuator.check_ender_switch_button() and not self.button_pressed:
            self.button_pressed = True

        if not self.measurer.actuator.check_ender_switch_button() and self.button_pressed:
            self.button_pressed = False

        if (
            self.button_pressed and (self.currentIndex() == 2) and
            (self.measurer.measurement_state == MeasurementState.READY)
        ):
            self.measurement_start()


def create_main_window(measurer):
    resistances_widget = QWidget()
    deviation_widget = QWidget()
    results_widget = QWidget()
    loadUi(r"resistances.ui", resistances_widget)
    loadUi(r"deviation_selection.ui", deviation_widget)
    loadUi(r"measurement_results.ui", results_widget)

    main_widget = StackedWidget(resistances_widget, results_widget, deviation_widget, measurer)
    main_window = MainWindow(main_widget)
    return main_window


def create_error_msg_window(error_msg):
    error_msg_widget = QWidget()
    loadUi(r"error.ui", error_msg_widget)
    error_msg_widget.error_msg.setText(error_msg)
    error_msg_widget.button_virtual.clicked.connect(lambda: error_msg_widget.close())
    if system() == 'Linux':
        error_msg_widget.setCursor(Qt.BlankCursor)
    return error_msg_widget
