from PyQt5.QtWidgets import QApplication
from PyQt5.QtWidgets import (qApp, QDesktopWidget, QLabel, QMainWindow, QMessageBox, QPushButton,
                             QVBoxLayout, QWidget)
from PyQt5.QtCore import pyqtSignal, QObject, Qt
import sys
import serial.tools.list_ports
from platform import system
from typing import List

from measurer import Measurer
from gui import create_main_window, create_error_msg_window

from functools import partial
import traceback


XIMC_USB_VID = 0x1CBE
XIMC_USB_PID = 0x7
RIKOHM_USB_VID = 0x1CBC
RIKOHM_USB_PID = 0x40

ERROR_MSG_TEXT_RU = 'При работе приложения произошла ошибка.\n\
Сфотографируйте сообщение с ошибкой и обратитесь в техподдержку (eyepoint@physlab.ru).'


def show_exception(args: List[str], msg_title: str, msg_text: str, exc: str = ""):
    """
    Function shows message box with error.
    :param args: arguments from command line;
    :param msg_title: title of message box;
    :param msg_text: message text;
    :param exc: text of exception.
    """

    msg = QMessageBox()
    msg.setWindowTitle(msg_title)
    start_msg_text = ERROR_MSG_TEXT_RU

    msg.setText(f"{start_msg_text}\n{msg_text}")
    if exc:
        msg.setInformativeText(str(exc)[-500:-1])
    msg.exec_()


class ExceptionHandler(QObject):
    """
    Class to handle unexpected errors.
    """

    exception_raised: pyqtSignal = pyqtSignal(str, str, str)

    def exception_hook(self, exc_type: Exception, exc_value: Exception, exc_traceback: "traceback"):
        """
        Method handles unexpected errors.
        :param exc_type: exception class;
        :param exc_value: exception instance;
        :param exc_traceback: traceback object.
        """

        traceback.print_exception(exc_type, exc_value, exc_traceback)
        traceback_text = "".join(traceback.format_exception(exc_type, exc_value, exc_traceback))
        self.exception_raised.emit("Error", str(exc_value), traceback_text)


class ErrorWindow(QMainWindow):
    """
    Window with error message with traceback.
    """

    def __init__(self, error: type, trace_back: str, args: List[str]):
        """
        :param error: exception;
        :param trace_back: text of traceback;
        :param args: command line arguments.
        """

        super().__init__()
        self.central_widget: QWidget = None
        self.english: bool = True if "--en" in args else False
        self.init_ui(error, trace_back)

        if system() == 'Linux':
            self.setCursor(Qt.BlankCursor)

    def init_ui(self, error: type, trace_back: str):
        """
        Method initializes widgets on window.
        :param error: exception;
        :param trace_back: text of traceback.
        """

        self.setWindowTitle("Error" if self.english else "Ошибка")
        self.central_widget: QWidget = QWidget()
        self.setCentralWidget(self.central_widget)
        layout = QVBoxLayout(self.central_widget)
        start_msg_text = ERROR_MSG_TEXT_RU
        layout.addWidget(QLabel(f"{start_msg_text}\n{str(error)}"))
        traceback_label = QLabel(trace_back[-500:-1])
        traceback_label.setStyleSheet("color: #8b0000;")
        layout.addWidget(traceback_label)
        exit_btn = QPushButton("OK", self)
        exit_btn.setFixedWidth(100)
        exit_btn.clicked.connect(qApp.quit)
        layout.addWidget(exit_btn, alignment=Qt.AlignHCenter)
        self.central_widget.setLayout(layout)
        center_point = QDesktopWidget().availableGeometry().center()
        qt_rectangle = self.frameGeometry()
        qt_rectangle.moveCenter(center_point)
        self.move(qt_rectangle.topLeft())


def start_error_app(app: QApplication, error: type, trace_back: str = "", args: List[str] = None):
    """
    Function creates window to show error message that occurred when application was launched.
    :param app: application;
    :param error: exception;
    :param trace_back: text of traceback;
    :param args: arguments from command line.
    """

    window = ErrorWindow(error, trace_back, args)
    window.show()
    app.exec_()


def get_uris():
    ports = serial.tools.list_ports.comports()
    actuator_port = None
    ohmmeter_port = None
    error_msg = ''
    for port in ports:
        # ximc
        if port.vid == XIMC_USB_VID and port.pid == XIMC_USB_PID:
            if actuator_port is None:
                actuator_port = port
            else:
                error_msg += 'Найдено несколько контроллеров двигателей XIMC.\n'
        # rikohm
        if port.vid == RIKOHM_USB_VID and port.pid == RIKOHM_USB_PID:
            if ohmmeter_port is None:
                ohmmeter_port = port
            else:
                error_msg += 'Найдено несколько измерителей сопротивлений Rikohm.\n'

    if actuator_port is None:
        error_msg += 'Не найден контроллер двигателей XIMC.\n'
    if ohmmeter_port is None:
        error_msg += 'Не найден измеритель сопротивлений Rikohm.\n'

    if error_msg != '':
        error_msg = 'ПО не может быть запущено из-за следующей ошибки:\n' + error_msg
        error_msg += '\nНайденные в системе устройства:\n'

        row_format = "{:<15} {:<6} {:<6} {:<8} {:<10} {}"
        error_msg += (row_format.format(
            "Порт", "VID", "PID", "S/N", "Располож.",
            "Производитель, устройство"
        )) + '\n'
        error_msg += ("------------------------------------------------------------------------------\n")
        for port in ports:
            port.vid = port.vid if port.pid is not None else 0
            port.pid = port.pid if port.pid is not None else 0
            port.serial_number = port.serial_number if port.serial_number is not None else 0
            port.location = port.location if port.location is not None else "N/A"
            port.manufacturer = port.manufacturer if port.manufacturer is not None else "N/A"
            port.product = port.product if port.product is not None else "N/A"
            port.description = port.description if port.description is not None else "N/A"

            device_name_string = "{} {}".format(port.manufacturer, port.description)

            device_name_string_max_len = 40
            if len(device_name_string) > device_name_string_max_len:
                device_name_string = device_name_string[:device_name_string_max_len] + "..."

            error_msg += (row_format.format(
                port.device,
                hex(port.vid), hex(port.pid), port.serial_number,
                port.location,
                device_name_string
                + '\n'
            ))
        error_msg += '\nСфотографируйте это сообщение и обратитесь в техническую поддержку: info@physlab.ru'
        return (None, None, error_msg)

    if system() == 'Windows':
        ohmmeter_uri = r"com:\\.\{}".format(ohmmeter_port.device)
        actuator_uri = r"xi-com:\\.\{}".format(actuator_port.device)
    if system() == 'Linux':
        ohmmeter_uri = r"com:{}".format(ohmmeter_port.device)
        actuator_uri = r"xi-com:{}".format(actuator_port.device)
    return (ohmmeter_uri, actuator_uri, None)


def main(args: List[str]) -> int:
    app = QApplication(sys.argv)
    try:
        ohmmeter_uri, actuator_uri, error_msg = get_uris()
        if error_msg is not None:
            # Show error message and the virtual mode button
            m = Measurer(virtual=True)
            error_msg_win = create_error_msg_window(error_msg)
            error_msg_win.show()
            app.exec()
        else:
            m = Measurer(ohmmeter_uri=ohmmeter_uri, actuator_uri=actuator_uri, virtual=False)

        main_window = create_main_window(m)  # <- load main GUI
        main_window.show()

        exceprion_handler = ExceptionHandler()
        sys.excepthook = exceprion_handler.exception_hook
        exceprion_handler.exception_raised.connect(partial(show_exception, args))
        return app.exec()

    except Exception as exc:
        traceback.print_exc()
        trace_back = "".join(traceback.format_exception(*sys.exc_info()))
        start_error_app(app, exc, trace_back, args)
        return 1


if __name__ == "__main__":
    sys.exit(main(sys.argv))
