import numpy as np
from actuator import Actuator
from ohmmeter import Ohmmeter, OhmmeterVirtual
import os
import tempfile
import re
import urllib
import time
from enum import Enum
from copy import deepcopy

ACCELERATION_DELAY_S = 1
WAIT_FOR_MEASUREMENT_PERIOD_S = 0.05


class MeasurementState(Enum):
    READY = 1
    REQUIRED = 2
    IN_PROGRESS = 3


class MeasurementStateSignal(Enum):
    FINISHED = 0
    TERMINATED = 1
    TIMEOUT = 2


class Measurer:
    def __init__(self, ohmmeter_uri="", actuator_uri="", virtual=False) -> None:
        if virtual:
            # use URI for virtual device when there is new urllib python3 API
            tempdir = tempfile.gettempdir() + "/testdevice.bin"
            if os.altsep:
                tempdir = tempdir.replace(os.sep, os.altsep)
            # urlparse build wrong path if scheme is not file
            uri = urllib.parse.urlunparse(
                urllib.parse.ParseResult(
                    scheme="file", netloc=None, path=tempdir, params=None, query=None, fragment=None
                )
            )
            actuator_uri = re.sub(r'^file', 'xi-emu', uri)

            self.actuator = Actuator(actuator_uri)
            self.ohmmeter = OhmmeterVirtual(ohmmeter_uri)
            self.ohmmeter.set_actuator_for_emulation(self.actuator)

        else:
            self.ohmmeter = Ohmmeter(ohmmeter_uri)
            self.actuator = Actuator(actuator_uri)

        self.connect_devices()

        self.virtual = virtual
        self.measurement_state = MeasurementState.READY
        self.measurements = []
        self.new_measurements = []
        self.measurements_per_revolution = self.get_measurements_per_revolution()
        self.timer = None
        self.t_start = None
        self.timeout_for_single_measurement = 10  # ms
        self.timeout = self.timeout_for_single_measurement * self.measurements_per_revolution / 1000 + 10  # s
        self.progressBar = None

    def connect_devices(self):
        self.ohmmeter.connect()
        self.actuator.connect()

    def disconnect_devices(self):
        self.ohmmeter.disconnect()
        self.actuator.disconnect()

    def set_nominal_resistance(self, res1, res2):
        self.nominal_resistance1 = res1
        self.nominal_resistance2 = res2

    def get_measurements_per_revolution(self):
        return int(self.actuator.get_measurements_per_revolution())

    def start(self):
        if self.measurement_state != MeasurementState.READY:
            raise RuntimeError('Cannot start new measurement')
            return
        self.measurement_state = MeasurementState.REQUIRED

        # TODO: do we need to calibrate? Probably yes
        self.ohmmeter.calibrate()

        self.actuator.zero()
        self.actuator.left()
        time.sleep(ACCELERATION_DELAY_S)

        # Movement started. Now we receive measurements
        self.ohmmeter.reset_err()
        self.new_measurements = []
        self.measurement_state = MeasurementState.IN_PROGRESS
        self.t_start = time.time()
        self.timer.start()

    def pick_measurements(self):
        self.progressBar.setValue(100*len(self.new_measurements)/self.measurements_per_revolution)

        if time.time() - self.t_start > self.timeout:
            self.terminate()
            # no measurements = no signals from ximc = motor probably disconnected
            if len(self.new_measurements) == 0:
                # display message that motor is disconnected
                self.timer.signal_emitter.measurement_state_signal.emit(MeasurementStateSignal.TIMEOUT)
            else:
                # god knows what happened, error
                raise RuntimeError("Time out. {} measurements done in {} s".format(
                        len(self.new_measurements),
                        time.time() - self.t_start
                    )
                )

        if len(self.new_measurements) < self.measurements_per_revolution:
            status = self.ohmmeter.status()
            self.ohmmeter.check_ohmmeter_error(status)

            n_left = self.measurements_per_revolution - len(self.new_measurements)
            for _ in range(min(status.buffer_length, n_left)):
                measurement = self.ohmmeter.get_one_measurement()
                self.new_measurements.append((
                    measurement[0] / 1000,
                    measurement[1] / 1000
                ))
        else:
            self.finish()

    def terminate(self):
        self.timer.stop()
        self.actuator.sstp()
        self.actuator.wait_for_stop(10)
        self.measurement_state = MeasurementState.READY
        self.timer.signal_emitter.measurement_state_signal.emit(MeasurementStateSignal.TERMINATED)

    def finish(self):
        self.timer.stop()
        self.actuator.sstp()
        self.actuator.wait_for_stop(10)
        self.measurement_state = MeasurementState.READY

        self.measurements = deepcopy(self.new_measurements)
        self.timer.signal_emitter.measurement_state_signal.emit(MeasurementStateSignal.FINISHED)

    def get_measurements(self):
        m = list(zip(*self.measurements))
        # Верх. канал = АЦП 2, нижн. канал = АЦП 1
        measurements_a = m[1]
        measurements_b = m[0]
        angles = np.linspace(0, 360, int(self.measurements_per_revolution), endpoint=False)
        return (angles, measurements_a, measurements_b)
