from actuator import Actuator
import numpy as np
from rikohm import RikohmDeviceHandle


class OhmmeterBase:
    def __init__(self, device_uri: str) -> None:
        pass

    def connect(self) -> None:
        raise NotImplementedError()

    def disconnect(self) -> None:
        raise NotImplementedError()

    def get_settings(self):
        raise NotImplementedError()

    def set_settings(self):
        raise NotImplementedError()

    def get_one_measurement(self):
        raise NotImplementedError()

    def reset_err(self) -> None:
        raise NotImplementedError()

    def status(self):
        raise NotImplementedError()

    def calibrate(self):
        raise NotImplementedError()


class Ohmmeter(OhmmeterBase):
    def __init__(self, device_uri: str) -> None:
        super().__init__(device_uri)
        self._device_uri = device_uri
        self.device = RikohmDeviceHandle(self._device_uri)
        self.band = None
        self.gain = None

    def connect(self) -> None:
        self.device.open_device()

    def disconnect(self) -> None:
        self.device.close_device()

    def get_one_measurement(self):
        resistances = self.device.get_last_buffer().resistances
        return resistances

    def reset_err(self) -> None:
        return self.device.reset_errors()

    def status(self):
        return self.device.get_status()

    def check_ohmmeter_error(self, ohmmeter_status):
        msg = ""
        if ohmmeter_status.errors == 0:
            return
        if ohmmeter_status.errors & 0x02:
            msg += "Measuremets overrun error.\n"
        if ohmmeter_status.errors & 0x01:
            msg += "Ohmmeter Buffer overflow error.\n"
        # TODO add other errors
        msg += "### Ext trg {}. Buf len: {}. Errors: {}".format(
            ohmmeter_status.external_input_enabled,
            ohmmeter_status.buffer_length,
            bin(ohmmeter_status.errors)
        )
        raise RuntimeError(msg)

    def calibrate(self):
        self.device.adc_internal_zero_calibration(1)
        self.device.adc_internal_fullscale_calibration(1)
        self.device.adc_internal_zero_calibration(2)
        self.device.adc_internal_fullscale_calibration(2)


class OhmmeterVirtual(OhmmeterBase):
    def __init__(self, device_uri: str) -> None:
        super().__init__(device_uri)
        self._device_uri = device_uri
        self._actuator = None
        self.nominal_resistance1 = 3300
        self.nominal_resistance2 = 100
        self.noise_factor = 0.008
        self.measurement_counter = 0

    def connect(self) -> None:
        pass

    def disconnect(self) -> None:
        pass

    def set_actuator_for_emulation(self, actuator: Actuator):
        self._actuator = actuator

    def get_one_measurement(self):
        measurements_per_revolution = 200
        if self._actuator is not None:
            measurements_per_revolution = self._actuator.get_measurements_per_revolution()

        self.measurement_counter += 1
        self.measurement_counter %= measurements_per_revolution

        angle = self.measurement_counter / measurements_per_revolution * 360
        resistance_rel = np.min((angle / 330, 1))

        resistance1 = self.nominal_resistance1 * resistance_rel
        resistance1 = self.add_noise(self.nominal_resistance1, resistance1)
        resistance2 = self.nominal_resistance2 * resistance_rel
        resistance2 = self.add_noise(self.nominal_resistance2, resistance2)
        return [resistance2 * 1000, resistance1 * 1000]

    def add_noise(self, max_value, value):
        noise_ampl = max_value * self.noise_factor
        value = value + noise_ampl * np.random.random()
        return value

    def reset_err(self) -> None:
        return

    def status(self):
        status = RikohmDeviceHandle.GetStatusResponse()
        status.errors = 0
        status.adc_busy = 0
        status.external_input_enabled = 1
        status.buffer_length = 1
        return status

    def check_ohmmeter_error(self, ohmmeter_status):
        return

    def calibrate(self):
        return
